package ecs

import (
	"github.com/hajimehoshi/ebiten/v2"
	"github.com/yohamta/donburi"
	"github.com/yohamta/donburi/filter"
	"github.com/yohamta/donburi/internal/component"
	"github.com/yohamta/donburi/query"
)

// LayerID is used to specify a layer.
type LayerID int

// ECS represents an entity-component-system.
type ECS struct {
	// World is the underlying world of the ECS.
	World donburi.World
	// Time manages the time of the world.
	Time *Time

	layers         []*Layer
	systems        []UpdateSystem
	startupSystems []UpdateSystem
}

// NewQuery creates a new query.
func NewQuery(l LayerID, f filter.LayoutFilter) *query.Query {
	layerFilter := filter.Contains(getLayer(l).tag)
	if f == nil {
		return query.NewQuery(layerFilter)
	}
	return query.NewQuery(filter.And(layerFilter, f))
}

// NewECS creates a new ECS with the specified world.
func NewECS(w donburi.World) *ECS {
	ecs := &ECS{
		World: w,
		Time:  NewTime(),

		systems:        []UpdateSystem{},
		layers:         []*Layer{},
		startupSystems: []UpdateSystem{},
	}

	return ecs
}

// AddSystems adds systems.
func (ecs *ECS) AddSystems(s ...System) *ECS {
	for _, ss := range s {
		ecs.AddSystem(ss)
	}
	return ecs
}

// AddSystem adds a system.
func (ecs *ECS) AddSystem(s System) *ECS {
	if s.Update != nil {
		ecs.addUpdateSystem(s.Update)
	}
	if s.Draw != nil {
		ecs.addDrawSystem(s.Layer, s.Draw)
	}
	return ecs
}

// Update runs systems
func (ecs *ECS) Update() {
	ecs.Time.Update()
	for _, sys := range ecs.systems {
		sys(ecs)
	}
}

// DrawLayer executes all draw systems of the specified layer.
func (ecs *ECS) DrawLayer(l LayerID, screen *ebiten.Image) {
	ecs.getLayer(l).draw(ecs, screen)
}

// Draw executes all draw systems.
func (ecs *ECS) Draw(screen *ebiten.Image) {
	for _, l := range ecs.layers {
		l.draw(ecs, screen)
	}
}

// Create creates a new entity
func (ecs *ECS) Create(l LayerID, components ...*donburi.ComponentType) donburi.Entity {
	entry := ecs.World.Entry(ecs.World.Create(components...))
	entry.AddComponent(ecs.getLayer(l).tag)
	return entry.Entity()
}

// Create creates a new entity
func (ecs *ECS) CreateMany(l LayerID, n int, components ...*component.ComponentType) []donburi.Entity {
	comps := append(components, ecs.getLayer(l).tag)
	return ecs.World.CreateMany(n, comps...)
}

// Pause pauses the world.
func (ecs *ECS) Pause() {
	ecs.Time.Pause()
}

// Resume resumes the world.
func (ecs *ECS) Resume() {
	ecs.Time.Resume()
}

// IsPaused returns a boolean value indicating whether the world is paused.
func (ecs *ECS) IsPaused() bool {
	return ecs.Time.isPaused
}

func (ecs *ECS) getLayer(layerID LayerID) *Layer {
	if int(layerID) >= len(ecs.layers) {
		ecs.layers = append(ecs.layers, make([]*Layer, int(layerID)-len(ecs.layers)+1)...)
	}
	if ecs.layers[layerID] == nil {
		ecs.layers[layerID] = newLayer(getLayer(layerID))
	}
	return ecs.layers[layerID]
}

func (ecs *ECS) addUpdateSystem(systems ...UpdateSystem) *ECS {
	for _, s := range systems {
		ecs.systems = append(ecs.systems, s)
	}
	return ecs
}

func (ecs *ECS) addDrawSystem(l LayerID, s DrawSystem) *ECS {
	ecs.getLayer(l).addDrawSystem(s)
	return ecs
}
